@@ -13,11 +13,10 @@
 #include <ngx_config.h>
 #include <ngx_core.h>
 #include <ngx_event.h>
+#include <ngx_event_connect.h>
 #include <njs.h>
 #include <njs_rbtree.h>
 #include <njs_arr.h>
-#include "ngx_js_fetch.h"
-#include "ngx_js_shared_dict.h"
 
 #if (NJS_HAVE_QUICKJS)
 #include <qjs.h>
@@ -41,6 +40,7 @@
 
 #define ngx_js_buffer_type(btype) ((btype) & ~NGX_JS_DEPRECATED)
 
+#if (NJS_HAVE_QUICKJS)
 /*
  * This static table solves the problem of a native QuickJS approach
  * which uses a static variables of type JSClassID and JS_NewClassID() to
@@ -49,20 +49,25 @@
  * are loaded dynamically.
  */
 
-#define NGX_QJS_CLASS_ID_OFFSET (QJS_CORE_CLASS_ID_LAST)
-#define NGX_QJS_CLASS_ID_CONSOLE (NGX_QJS_CLASS_ID_OFFSET + 1)
-#define NGX_QJS_CLASS_ID_HTTP_REQUEST (NGX_QJS_CLASS_ID_OFFSET + 2)
-#define NGX_QJS_CLASS_ID_HTTP_PERIODIC (NGX_QJS_CLASS_ID_OFFSET + 3)
-#define NGX_QJS_CLASS_ID_HTTP_VARS (NGX_QJS_CLASS_ID_OFFSET + 4)
-#define NGX_QJS_CLASS_ID_HTTP_HEADERS_IN (NGX_QJS_CLASS_ID_OFFSET + 5)
-#define NGX_QJS_CLASS_ID_HTTP_HEADERS_OUT (NGX_QJS_CLASS_ID_OFFSET + 6)
-#define NGX_QJS_CLASS_ID_STREAM_SESSION (NGX_QJS_CLASS_ID_OFFSET + 7)
-#define NGX_QJS_CLASS_ID_STREAM_PERIODIC (NGX_QJS_CLASS_ID_OFFSET + 8)
-#define NGX_QJS_CLASS_ID_STREAM_FLAGS (NGX_QJS_CLASS_ID_OFFSET + 9)
-#define NGX_QJS_CLASS_ID_STREAM_VARS (NGX_QJS_CLASS_ID_OFFSET + 10)
-#define NGX_QJS_CLASS_ID_SHARED (NGX_QJS_CLASS_ID_OFFSET + 11)
-#define NGX_QJS_CLASS_ID_SHARED_DICT (NGX_QJS_CLASS_ID_OFFSET + 12)
-#define NGX_QJS_CLASS_ID_SHARED_DICT_ERROR (NGX_QJS_CLASS_ID_OFFSET + 13)
+enum {
+    NGX_QJS_CLASS_ID_CONSOLE = QJS_CORE_CLASS_ID_LAST,
+    NGX_QJS_CLASS_ID_HTTP_REQUEST,
+    NGX_QJS_CLASS_ID_HTTP_PERIODIC,
+    NGX_QJS_CLASS_ID_HTTP_VARS,
+    NGX_QJS_CLASS_ID_HTTP_HEADERS_IN,
+    NGX_QJS_CLASS_ID_HTTP_HEADERS_OUT,
+    NGX_QJS_CLASS_ID_STREAM_SESSION,
+    NGX_QJS_CLASS_ID_STREAM_PERIODIC,
+    NGX_QJS_CLASS_ID_STREAM_FLAGS,
+    NGX_QJS_CLASS_ID_STREAM_VARS,
+    NGX_QJS_CLASS_ID_SHARED,
+    NGX_QJS_CLASS_ID_SHARED_DICT,
+    NGX_QJS_CLASS_ID_SHARED_DICT_ERROR,
+    NGX_QJS_CLASS_ID_FETCH_HEADERS,
+    NGX_QJS_CLASS_ID_FETCH_REQUEST,
+    NGX_QJS_CLASS_ID_FETCH_RESPONSE,
+};
+#endif
 
 
 typedef struct ngx_js_loc_conf_s ngx_js_loc_conf_t;
@@ -76,9 +81,7 @@ typedef ngx_pool_t *(*ngx_external_pool_pt)(njs_external_ptr_t e);
 typedef void (*ngx_js_event_finalize_pt)(njs_external_ptr_t e, ngx_int_t rc);
 typedef ngx_resolver_t *(*ngx_external_resolver_pt)(njs_external_ptr_t e);
 typedef ngx_msec_t (*ngx_external_timeout_pt)(njs_external_ptr_t e);
-typedef ngx_flag_t (*ngx_external_flag_pt)(njs_external_ptr_t e);
-typedef ngx_flag_t (*ngx_external_size_pt)(njs_external_ptr_t e);
-typedef ngx_ssl_t *(*ngx_external_ssl_pt)(njs_external_ptr_t e);
+typedef ngx_js_loc_conf_t *(*ngx_js_external_loc_conf_pt)(njs_external_ptr_t e);
 typedef ngx_js_ctx_t *(*ngx_js_external_ctx_pt)(njs_external_ptr_t e);
 
 
@@ -121,6 +124,7 @@ typedef struct {
     ngx_uint_t             type;                                              \
     ngx_engine_t          *engine;                                            \
     ngx_uint_t             reuse;                                             \
+    size_t                 reuse_max_size;                                    \
     ngx_js_queue_t        *reuse_queue;                                       \
     ngx_str_t              cwd;                                               \
     ngx_array_t           *imports;                                           \
@@ -130,10 +134,34 @@ typedef struct {
                                                                               \
     size_t                 buffer_size;                                       \
     size_t                 max_response_body_size;                            \
-    ngx_msec_t             timeout
+    ngx_msec_t             timeout;                                           \
+                                                                              \
+    ngx_uint_t             fetch_keepalive;                                   \
+    ngx_uint_t             fetch_keepalive_requests;                          \
+    ngx_msec_t             fetch_keepalive_time;                              \
+    ngx_msec_t             fetch_keepalive_timeout;                           \
+    ngx_queue_t            fetch_keepalive_cache;                             \
+    ngx_queue_t            fetch_keepalive_free;                              \
+                                                                              \
+    ngx_url_t              *fetch_proxy_url;                                  \
+    ngx_str_t               fetch_proxy_auth_header;                          \
+                                                                              \
+    ngx_int_t             (*eval_proxy_url)(ngx_pool_t *pool,                 \
+                                            void *request,                    \
+                                            void *module_conf,                \
+                                            ngx_url_t **url_out,              \
+                                            ngx_str_t *auth_out)
+
+#define ngx_js_conf_dynamic_proxy(conf)                                       \
+     ((conf)->eval_proxy_url != NULL)
+
+#define ngx_js_conf_proxy(conf)                                               \
+    (((conf)->fetch_proxy_url != NULL                                         \
+      && (conf)->fetch_proxy_url->host.len > 0)                               \
+     || ngx_js_conf_dynamic_proxy(conf))
 
 
-#if defined(NGX_HTTP_SSL) || defined(NGX_STREAM_SSL)
+#if (NGX_SSL)
 #define NGX_JS_COMMON_LOC_CONF                                                \
     _NGX_JS_COMMON_LOC_CONF;                                                  \
                                                                               \
@@ -184,8 +212,10 @@ struct ngx_js_loc_conf_s {
 
 
 typedef struct {
-    ngx_str_t fname;
-    unsigned  flags;
+    ngx_str_t    fname;
+    unsigned     flags;
+    u_char      *file_name;
+    ngx_uint_t   line;
 } ngx_js_set_t;
 
 
@@ -265,31 +295,35 @@ struct ngx_engine_s {
 };
 
 
+enum {
+    NGX_JS_EXTERNAL_CONNECTION = 0,
+    NGX_JS_EXTERNAL_POOL,
+    NGX_JS_EXTERNAL_RESOLVER,
+    NGX_JS_EXTERNAL_RESOLVER_TIMEOUT,
+    NGX_JS_EXTERNAL_EVENT_FINALIZE,
+    NGX_JS_EXTERNAL_LOC_CONF,
+    NGX_JS_EXTERNAL_CTX,
+    NGX_JS_EXTERNAL_MAIN_CONF,
+};
+
 #define ngx_external_connection(vm, e)                                        \
-    (*((ngx_connection_t **) ((u_char *) (e) + njs_vm_meta(vm, 0))))
+    (*((ngx_connection_t **)                                                  \
+       ((u_char *) (e) + njs_vm_meta(vm, NGX_JS_EXTERNAL_CONNECTION))))
 #define ngx_external_pool(vm, e)                                              \
-    ((ngx_external_pool_pt) njs_vm_meta(vm, 1))(e)
+    ((ngx_external_pool_pt) njs_vm_meta(vm, NGX_JS_EXTERNAL_POOL))(e)
 #define ngx_external_resolver(vm, e)                                          \
-    ((ngx_external_resolver_pt) njs_vm_meta(vm, 2))(e)
+    ((ngx_external_resolver_pt) njs_vm_meta(vm, NGX_JS_EXTERNAL_RESOLVER))(e)
 #define ngx_external_resolver_timeout(vm, e)                                  \
-    ((ngx_external_timeout_pt) njs_vm_meta(vm, 3))(e)
+    ((ngx_external_timeout_pt)                                                \
+      njs_vm_meta(vm, NGX_JS_EXTERNAL_RESOLVER_TIMEOUT))(e)
 #define ngx_external_event_finalize(vm) \
-    ((ngx_js_event_finalize_pt) njs_vm_meta(vm, 4))
-#define ngx_external_ssl(vm, e)                                               \
-    ((ngx_external_ssl_pt) njs_vm_meta(vm, 5))(e)
-#define ngx_external_ssl_verify(vm, e)                                        \
-    ((ngx_external_flag_pt) njs_vm_meta(vm, 6))(e)
-#define ngx_external_fetch_timeout(vm, e)                                     \
-    ((ngx_external_timeout_pt) njs_vm_meta(vm, 7))(e)
-#define ngx_external_buffer_size(vm, e)                                       \
-    ((ngx_external_size_pt) njs_vm_meta(vm, 8))(e)
-#define ngx_external_max_response_buffer_size(vm, e)                          \
-    ((ngx_external_size_pt) njs_vm_meta(vm, 9))(e)
-#define NGX_JS_MAIN_CONF_INDEX  10
-#define ngx_main_conf(vm)                                                     \
-	((ngx_js_main_conf_t *) njs_vm_meta(vm, NGX_JS_MAIN_CONF_INDEX))
+    ((ngx_js_event_finalize_pt) njs_vm_meta(vm, NGX_JS_EXTERNAL_EVENT_FINALIZE))
+#define ngx_external_loc_conf(vm, e)                                          \
+    ((ngx_js_external_loc_conf_pt) njs_vm_meta(vm, NGX_JS_EXTERNAL_LOC_CONF))(e)
 #define ngx_external_ctx(vm, e) \
-    ((ngx_js_external_ctx_pt) njs_vm_meta(vm, 11))(e)
+    ((ngx_js_external_ctx_pt) njs_vm_meta(vm, NGX_JS_EXTERNAL_CTX))(e)
+#define ngx_main_conf(vm)                                                     \
+	((ngx_js_main_conf_t *) njs_vm_meta(vm, NGX_JS_EXTERNAL_MAIN_CONF))
 
 
 #define ngx_js_prop(vm, type, value, start, len)                              \
@@ -310,10 +344,13 @@ void ngx_js_ctx_init(ngx_js_ctx_t *ctx, ngx_log_t *log);
 void ngx_js_ctx_destroy(ngx_js_ctx_t *ctx, ngx_js_loc_conf_t *conf);
 ngx_int_t ngx_js_call(njs_vm_t *vm, njs_function_t *func,
     njs_opaque_value_t *args, njs_uint_t nargs);
-ngx_int_t ngx_js_exception(njs_vm_t *vm, ngx_str_t *s);
+ngx_int_t ngx_js_log_exception(njs_vm_t *vm, ngx_log_t *log, const char *txt);
 ngx_engine_t *ngx_njs_clone(ngx_js_ctx_t *ctx, ngx_js_loc_conf_t *cf,
     void *external);
 
+#define NGX_CHB_CTX_INIT(chain, pool)                                        \
+    njs_chb_init(chain, pool, (njs_chb_alloc_t) ngx_palloc, NULL)
+
 #if (NJS_HAVE_QUICKJS)
 
 typedef struct ngx_qjs_event_s ngx_qjs_event_t;
@@ -342,9 +379,14 @@ void ngx_engine_qjs_destroy(ngx_engine_t *e, ngx_js_ctx_t *ctx,
     ngx_js_loc_conf_t *conf);
 ngx_int_t ngx_qjs_call(JSContext *cx, JSValue function, JSValue *argv,
     int argc);
-ngx_int_t ngx_qjs_exception(ngx_engine_t *e, ngx_str_t *s);
+ngx_int_t ngx_qjs_log_exception(ngx_engine_t *e, ngx_log_t *log,
+    const char *txt);
 ngx_int_t ngx_qjs_integer(JSContext *cx, JSValueConst val, ngx_int_t *n);
-ngx_int_t ngx_qjs_string(JSContext *cx, JSValueConst val, ngx_str_t *str);
+ngx_int_t ngx_qjs_string(JSContext *cx, ngx_pool_t *pool, JSValueConst val,
+    ngx_str_t *dst);
+
+JSValue ngx_qjs_ext_fetch(JSContext *cx, JSValueConst this_val, int argc,
+     JSValueConst *argv);
 
 #define ngx_qjs_prop(cx, type, start, len)                                   \
     ((type == NGX_JS_STRING) ? qjs_string_create(cx, start, len)             \
@@ -353,35 +395,40 @@ ngx_int_t ngx_qjs_string(JSContext *cx, JSValueConst val, ngx_str_t *str);
 #define ngx_qjs_meta(cx, i)                                                  \
     ((uintptr_t *) JS_GetRuntimeOpaque(JS_GetRuntime(cx)))[i]
 #define ngx_qjs_external_connection(cx, e)                                   \
-    (*((ngx_connection_t **) ((u_char *) (e) + ngx_qjs_meta(cx, 0))))
+    (*((ngx_connection_t **)                                                 \
+       ((u_char *) (e) + ngx_qjs_meta(cx, NGX_JS_EXTERNAL_CONNECTION))))
 #define ngx_qjs_external_pool(cx, e)                                         \
-    ((ngx_external_pool_pt) ngx_qjs_meta(cx, 1))(e)
+    ((ngx_external_pool_pt) ngx_qjs_meta(cx, NGX_JS_EXTERNAL_POOL))(e)
 #define ngx_qjs_external_resolver(cx, e)                                     \
-    ((ngx_external_resolver_pt) ngx_qjs_meta(cx, 2))(e)
+    ((ngx_external_resolver_pt) ngx_qjs_meta(cx, NGX_JS_EXTERNAL_RESOLVER))(e)
 #define ngx_qjs_external_resolver_timeout(cx, e)                             \
-    ((ngx_external_timeout_pt) ngx_qjs_meta(cx, 3))(e)
+    ((ngx_external_timeout_pt)                                               \
+     ngx_qjs_meta(cx, NGX_JS_EXTERNAL_RESOLVER_TIMEOUT))(e)
 #define ngx_qjs_external_event_finalize(cx)                                  \
-    ((ngx_js_event_finalize_pt) ngx_qjs_meta(cx, 4))
+    ((ngx_js_event_finalize_pt) ngx_qjs_meta(cx, NGX_JS_EXTERNAL_EVENT_FINALIZE))
+#define ngx_qjs_external_loc_conf(cx, e)                                     \
+    ((ngx_js_external_loc_conf_pt) ngx_qjs_meta(cx, NGX_JS_EXTERNAL_LOC_CONF))(e)
 #define ngx_qjs_external_ssl(cx, e)                                          \
-    ((ngx_external_ssl_pt) ngx_qjs_meta(cx, 5))(e)
+    (ngx_qjs_external_loc_conf(cx, e)->ssl)
 #define ngx_qjs_external_ssl_verify(cx, e)                                   \
-    ((ngx_external_flag_pt) ngx_qjs_meta(cx, 6))(e)
+    (ngx_qjs_external_loc_conf(cx, e)->ssl_verify)
 #define ngx_qjs_external_fetch_timeout(cx, e)                                \
-    ((ngx_external_timeout_pt) ngx_qjs_meta(cx, 7))(e)
+    (ngx_qjs_external_loc_conf(cx, e)->timeout)
 #define ngx_qjs_external_buffer_size(cx, e)                                  \
-    ((ngx_external_size_pt) ngx_qjs_meta(cx, 8))(e)
+    (ngx_qjs_external_loc_conf(cx, e)->buffer_size)
 #define ngx_qjs_external_max_response_buffer_size(cx, e)                     \
-    ((ngx_external_size_pt) ngx_qjs_meta(cx, 9))(e)
-#define ngx_qjs_main_conf(cx)                                                \
-    ((ngx_js_main_conf_t *) ngx_qjs_meta(cx, NGX_JS_MAIN_CONF_INDEX))
+    (ngx_qjs_external_loc_conf(cx, e)->max_response_body_size)
 #define ngx_qjs_external_ctx(cx, e)                                          \
-    ((ngx_js_external_ctx_pt) ngx_qjs_meta(cx, 11))(e)
+    ((ngx_js_external_ctx_pt) ngx_qjs_meta(cx, NGX_JS_EXTERNAL_CTX))(e)
+#define ngx_qjs_main_conf(cx)                                                \
+    ((ngx_js_main_conf_t *) ngx_qjs_meta(cx, NGX_JS_EXTERNAL_MAIN_CONF))
 
 extern qjs_module_t  qjs_webcrypto_module;
 extern qjs_module_t  qjs_xml_module;
 extern qjs_module_t  qjs_zlib_module;
 extern qjs_module_t  ngx_qjs_ngx_module;
 extern qjs_module_t  ngx_qjs_ngx_shared_dict_module;
+extern qjs_module_t  ngx_qjs_ngx_fetch_module;
 
 #endif
 
@@ -394,6 +441,9 @@ void ngx_js_logger(ngx_connection_t *c, ngx_uint_t level,
 char * ngx_js_import(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
 char * ngx_js_engine(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
 char * ngx_js_preload_object(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
+char * ngx_js_fetch_proxy(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
+ngx_int_t ngx_js_parse_proxy_url(ngx_pool_t *pool, ngx_log_t *log,
+    ngx_str_t *url_str, ngx_url_t **url_out, ngx_str_t *auth_header_out);
 ngx_int_t ngx_js_merge_vm(ngx_conf_t *cf, ngx_js_loc_conf_t *conf,
     ngx_js_loc_conf_t *prev,
     ngx_int_t (*init_vm)(ngx_conf_t *cf, ngx_js_loc_conf_t *conf));
@@ -415,6 +465,7 @@ njs_int_t ngx_js_ext_flags(njs_vm_t *vm, njs_object_prop_t *prop, uint32_t unuse
     njs_value_t *value, njs_value_t *setval, njs_value_t *retval);
 
 ngx_int_t ngx_js_string(njs_vm_t *vm, njs_value_t *value, njs_str_t *str);
+ngx_int_t ngx_js_ngx_string(njs_vm_t *vm, njs_value_t *value, ngx_str_t *str);
 ngx_int_t ngx_js_integer(njs_vm_t *vm, njs_value_t *value, ngx_int_t *n);
 const char *ngx_js_errno_string(int errnum);
 
@@ -429,4 +480,7 @@ extern njs_module_t  njs_xml_module;
 extern njs_module_t  njs_zlib_module;
 
 
+#include "ngx_js_fetch.h"
+#include "ngx_js_shared_dict.h"
+
 #endif /* _NGX_JS_H_INCLUDED_ */
